import logging
import traceback

from jet_django.deps.rest_framework import status
from jet_django.deps.rest_framework.views import exception_handler, set_rollback
from jet_django.deps.rest_framework.response import Response

logger = logging.getLogger(__name__)


def jet_exception_handler(exc, context):
    response = exception_handler(exc, context)

    if not response:
        data = {'detail': 'Server Error'}

        set_rollback()
        response = Response(data, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        logger.exception('Jet view exception', exc_info=exc)
        traceback.print_exc()

    if context and 'request' in context and context['request'].method == 'OPTIONS':
        response.status_code = 204

        response['Access-Control-Max-Age'] = '1728000'
        response['Content-Type'] = 'text/plain; charset=utf-8'
        response['Content-Length'] = 0

    return response



from jet_django.admin.jet import jet
from jet_django.deps.rest_framework import serializers


class MessageSerializer(serializers.Serializer):
    name = serializers.CharField()
    params = serializers.JSONField(required=False, default={})

    def save(self, **kwargs):
        handler = jet.get_message_handler(self.validated_data['name'])

        if not handler:
            return

        return handler(self.validated_data.get('params'))



from django.conf import settings

# django-crispy-forms is optional
try:
    import crispy_forms
except ImportError:
    crispy_forms = None


def is_crispy():
    return 'crispy_forms' in settings.INSTALLED_APPS and crispy_forms


# coreapi is optional (Note that uritemplate is a dependency of coreapi)
# Fixes #525 - cannot simply import from jet_django.deps.rest_framework.compat, due to
# import issues w/ django-guardian.
try:
    import coreapi
except ImportError:
    coreapi = None

try:
    import coreschema
except ImportError:
    coreschema = None



from django.db.models import Case, When, Value, IntegerField

from jet_django.deps.rest_framework import serializers


def reset_order_serializer_factory(build_queryset):
    class ResetOrderSerializer(serializers.Serializer):
        ordering_field = serializers.CharField()
        ordering = serializers.CharField(required=False, allow_null=True)
        value_ordering = serializers.CharField(required=False, allow_null=True)

        def save(self):
            ordering_field = self.validated_data['ordering_field']
            ordering = self.validated_data.get('ordering')
            value_ordering = self.validated_data.get('value_ordering')

            qs = build_queryset
            order_by = []

            if value_ordering:
                field, values_str = value_ordering.split('-', 2)
                values = values_str.split(',')
                qs = qs.annotate(
                    __custom_order__=Case(
                        *[When(**dict([(field, x), ('then', Value(i))])) for i, x in enumerate(values)],
                        default=Value(len(values)),
                        output_field=IntegerField()
                    )
                )
                order_by.append('__custom_order__')

            if ordering:
                order_by.extend(ordering.split(','))

            if len(order_by):
                qs = qs.order_by(*order_by)

            i = 1
            for instance in qs:
                setattr(instance, ordering_field, i)
                instance.save(update_fields=[ordering_field])
                i += 1

    return ResetOrderSerializer



from django.db import connection


def escape_sql_identifier(identifier):
    if connection.vendor == 'mysql':
        return '`{}`'.format(identifier)
    else:
        return '"{}"'.format(identifier)



from jet_django.deps import django_filters


class GEOSGeometryFilter(django_filters.CharFilter):

    def filter(self, qs, value):
        try:
            from django.contrib.gis.geos import GEOSGeometry
            value = GEOSGeometry(value)
            return super().filter(qs, value)
        except (ValueError, TypeError, ImportError):
            return qs



from django.conf.urls import url
from jet_django.deps.rest_framework.routers import DefaultRouter

from jet_django.admin.jet import jet
from jet_django.views.file_upload import FileUploadView
from jet_django.views.model_description import ModelDescriptionView
from jet_django.views.register import RegisterView
from jet_django.views.root import RootView
from jet_django.views.sql import SqlView
from jet_django.views.message import MessageView

app_name = 'jet_django'


def init_urls():
    class Router(DefaultRouter):
        include_root_view = False

    router = Router()

    jet.register_related_models()

    for model in jet.models:
        router.register(model.viewset_url, model.viewset)

    extra_urls = [
        url(r'^model_descriptions/', ModelDescriptionView.as_view(), name='model-descriptions'),
        url(r'^register/', RegisterView.as_view(), name='register'),
        url(r'^sql/', SqlView.as_view(), name='sql'),
        url(r'^file_upload/', FileUploadView.as_view(), name='file-upload'),
        url(r'^messages/', MessageView.as_view(), name='message'),
        url(r'^$', RootView.as_view(), name='root')
    ]

    api_urls = router.urls + extra_urls

    return api_urls


jet_urls = init_urls()
urlpatterns = jet_urls



import logging

import sys
from django.apps import AppConfig
from django.apps import apps
from django.db import ProgrammingError

from jet_django import settings

logger = logging.getLogger('jet_django')


class JetDjangoConfig(AppConfig):
    name = 'jet_django'

    def check_token(self):
        from jet_django.utils.backend import register_token, is_token_activated

        is_command = len(sys.argv) > 1 and sys.argv[1].startswith('jet_')

        if not is_command and settings.JET_REGISTER_TOKEN_ON_START:
            try:
                print('[JET] Checking if token is not activated yet...')
                token, created = register_token()

                if not token:
                    return

                if not is_token_activated(token):
                    print('[!] Your server token is not activated')
                    print('[!] Token: {}'.format(token.token))
                else:
                    print('[JET] Token activated')
            except ProgrammingError as e:
                no_migrations = str(e).find('relation "jet_django_token" does not exist') != -1
                if no_migrations:
                    print('[JET] Apply migrations first: python manage.py migrate jet_django')
                else:
                    print(e)
            except Exception as e:  # if no migrations yet
                print(e)
                pass

    def register_models(self):
        from jet_django.admin.jet import jet

        try:
            models = apps.get_models()

            for model in models:
                jet.register(model)
        except:  # if no migrations yet
            pass

    def ready(self):
        self.check_token()
        self.register_models()



from jet_django.deps.rest_framework import serializers


def model_detail_serializer_factory(build_model, build_fields):
    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = build_model
            fields = build_fields + ['__str__']

    return Serializer



from django.db import transaction
from django.db.models import F
from jet_django.deps.rest_framework import serializers, relations


def reorder_serializer_factory(build_queryset):
    class ReorderSerializer(serializers.Serializer):
        ordering_field = serializers.CharField()
        forward = serializers.BooleanField()
        segment_from = serializers.IntegerField()
        segment_to = serializers.IntegerField()
        item = relations.PrimaryKeyRelatedField(queryset=build_queryset)
        segment_by_ordering_field = serializers.BooleanField(default=False)

        def save(self):
            ordering_field = self.validated_data['ordering_field']

            if self.validated_data.get('segment_by_ordering_field'):
                segment_from = self.validated_data['segment_from']
                segment_to = self.validated_data['segment_to']
            else:
                segment_from_instance = build_queryset.get(pk=self.validated_data['segment_from'])
                segment_to_instance = build_queryset.get(pk=self.validated_data['segment_to'])

                segment_from = getattr(segment_from_instance, ordering_field)
                segment_to = getattr(segment_to_instance, ordering_field)

            with transaction.atomic():
                if self.validated_data['forward']:
                    build_queryset.filter(
                        **{
                            '{}__gte'.format(ordering_field): segment_from,
                            '{}__lte'.format(ordering_field): segment_to
                        }
                    ).update(
                        **{ordering_field: F(ordering_field) - 1}
                    )
                    build_queryset.filter(
                        pk=self.validated_data['item'].pk
                    ).update(
                        **{ordering_field: segment_to}
                    )
                else:
                    build_queryset.filter(
                        **{
                            '{}__gte'.format(ordering_field): segment_from,
                            '{}__lte'.format(ordering_field): segment_to
                        }
                    ).update(
                        **{ordering_field: F(ordering_field) + 1}
                    )
                    build_queryset.filter(
                        pk=self.validated_data['item'].pk
                    ).update(
                        **{ordering_field: segment_from}
                    )

    return ReorderSerializer



from jet_django.deps.rest_framework import views
from jet_django.deps.rest_framework.response import Response
from jet_django.mixins.cors_api_view import CORSAPIViewMixin
from jet_django.permissions import HasProjectPermissions
from jet_django.serializers.message import MessageSerializer

from django.http.response import HttpResponseBase


class MessageView(CORSAPIViewMixin, views.APIView):
    authentication_classes = ()
    permission_classes = (HasProjectPermissions,)

    def post(self, request, *args, **kwargs):
        serializer = MessageSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        result = serializer.save()

        if isinstance(result, HttpResponseBase):
            return result
        else:
            return Response(result)



from django.db import connection
from django.db.models.sql import Query
from django.db.models.sql.datastructures import BaseTable
from django.db.models.sql.compiler import SQLCompiler


def get_row_number(Model, instance, join_sql, join_args, where_sql, where_args, order_by_sql):
    pk = Model._meta.pk.name
    table = Model._meta.db_table

    with connection.cursor() as cursor:
        query = '''
            SELECT 
              __inner.__inner__row 
            FROM 
              (
                SELECT 
                  {}.{} as __inner__pk, 
                  ROW_NUMBER() OVER({}) AS __inner__row
                FROM 
                  {}
                {}
                {}
              ) AS __inner 
            WHERE 
              __inner.__inner__pk = %s
        '''.format(
            table,
            pk,
            order_by_sql,
            table,
            join_sql,
            where_sql
        )
        args = join_args + where_args + [instance.pk]
        cursor.execute(query, args)
        row = cursor.fetchone()

    if not row:
        return

    return row[0]


def get_row_siblings(Model, row_number, join_sql, join_args, where_sql, where_args, order_by_sql):
    pk = Model._meta.pk.name
    table = Model._meta.db_table

    has_prev = row_number > 1
    offset = row_number - 2 if has_prev else row_number - 1
    limit = 3 if has_prev else 2

    with connection.cursor() as cursor:
        query = '''
                    SELECT {}.{} 
                    FROM {} 
                    {}
                    {}
                    {} 
                    LIMIT %s 
                    OFFSET %s
                '''.format(
            table,
            pk,
            table,
            join_sql,
            where_sql,
            order_by_sql
        )
        args = join_args + where_args + [limit, offset]
        cursor.execute(query, args)
        rows = cursor.fetchall()

    if has_prev:
        next_index = 2
    else:
        next_index = 1

    if next_index >= len(rows):
        next_index = None

    if has_prev:
        prev_index = 0
    else:
        prev_index = None

    def map_row(row):
        columns = (x.name for x in cursor.description)
        return dict(zip(columns, row))

    return {
        'prev': map_row(rows[prev_index]) if prev_index is not None else None,
        'next': map_row(rows[next_index]) if next_index is not None else None
    }


def create_joins(queryset, compiler):
    join_queries = []
    join_args = []

    for key, value in queryset.query.alias_map.items():
        if isinstance(value, BaseTable):
            continue
        else:
            query, args = value.as_sql(compiler, connection)
            join_queries.append(query)
            join_args.extend(args)

    join_sql = ' '.join(join_queries)

    return join_sql, join_args


def create_where(queryset, compiler):
    where_query, where_args = queryset.query.where.as_sql(compiler, connection)
    return 'WHERE {}'.format(where_query) if where_query != '' else '', where_args


def create_order_by(compiler):
    compiler_order_by = compiler.get_order_by()

    if not compiler_order_by:
        compiler_order_by = []

    ordering = []
    for _, (o_sql, o_params, _) in compiler_order_by:
        ordering.append(o_sql)

    return 'ORDER BY %s' % ', '.join(ordering)


def get_model_siblings(Model, instance, queryset):
    pk = Model._meta.pk.name
    ordering = queryset.query.order_by

    if len(ordering) == 0 and len(Model._meta.ordering):
        ordering = Model._meta.ordering

    if not any(map(lambda x: x == pk or x == '-{}'.format(pk), ordering)):
        ordering = list(ordering) + ['-{}'.format(pk)]

    queryset.query.order_by = ordering
    compiler = SQLCompiler(queryset.query, connection, queryset.db)
    compiler.as_sql()

    join_sql, join_args = create_joins(queryset, compiler)
    where_sql, where_args = create_where(queryset, compiler)
    order_by_sql = create_order_by(compiler)

    row_number = get_row_number(Model, instance, join_sql, join_args, where_sql, where_args, order_by_sql)

    if not row_number:
        return {}

    return get_row_siblings(Model, row_number, join_sql, join_args, where_sql, where_args, order_by_sql)



# flake8: noqa
import pkgutil

from .filterset import FilterSet
from .filters import *

# We make the `rest_framework` module available without an additional import.
#   If DRF is not installed, no-op.
if pkgutil.find_loader('jet_django.deps.rest_framework') is not None:
    from . import rest_framework
del pkgutil

__version__ = '2.0.0'


def parse_version(version):
    '''
    '0.1.2.dev1' -> (0, 1, 2, 'dev1')
    '0.1.2' -> (0, 1, 2)
    '''
    v = version.split('.')
    ret = []
    for p in v:
        if p.isdigit():
            ret.append(int(p))
        else:
            ret.append(p)
    return tuple(ret)

VERSION = parse_version(__version__)
